@@ -1,15 +1,15 @@
 /* display.h -- How the display in Info is done.
-   $Id: display.h,v 1.1.1.1 2003/12/06 01:32:35 karl Exp $
+   $Id: display.h,v 1.7 2008/05/10 14:39:04 gray Exp $
 
    This file is part of GNU Info, a program for reading online documentation
    stored in Info format.
 
-   Copyright (C) 1993, 97 Free Software Foundation, Inc.
+   Copyright (C) 1993, 1997, 2004, 2007 Free Software Foundation, Inc.
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,8 +17,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
    Written by Brian Fox (bfox@ai.mit.edu). */
 
@@ -45,33 +44,39 @@ extern int display_inhibited;
 extern int display_was_interrupted_p;
 
 /* Initialize THE_DISPLAY to WIDTH and HEIGHT, with nothing in it. */
-extern void display_initialize_display ();
+extern void display_initialize_display (int width, int height);
 
 /* Clear all of the lines in DISPLAY making the screen blank. */
-extern void display_clear_display ();
+extern void display_clear_display (DISPLAY_LINE **display);
 
 /* Update the windows pointed to by WINDOWS in THE_DISPLAY.  This actually
    writes the text on the screen. */
-extern void display_update_display ();
+extern void display_update_display (WINDOW *window);
 
 /* Display WIN on THE_DISPLAY.  Unlike display_update_display (), this
    function only does one window. */
-extern void display_update_one_window ();
+extern void display_update_one_window (WINDOW *win);
 
 /* Move the screen cursor to directly over the current character in WINDOW. */
-extern void display_cursor_at_point ();
+extern void display_cursor_at_point (WINDOW *window);
 
 /* Scroll the region of the_display starting at START, ending at END, and
    moving the lines AMOUNT lines.  If AMOUNT is less than zero, the lines
    are moved up in the screen, otherwise down.  Actually, it is possible
    for no scrolling to take place in the case that the terminal doesn't
    support it.  This doesn't matter to us. */
-extern void display_scroll_display ();
+extern void display_scroll_display (int start, int end, int amount);
 
 /* Try to scroll lines in WINDOW.  OLD_PAGETOP is the pagetop of WINDOW before
    having had its line starts recalculated.  OLD_STARTS is the list of line
    starts that used to appear in this window.  OLD_COUNT is the number of lines
    that appear in the OLD_STARTS array. */
-extern void display_scroll_line_starts ();
+extern void display_scroll_line_starts (WINDOW *window, int old_pagetop,
+    char **old_starts, int old_count);
+
+extern size_t process_node_text
+        (WINDOW *win, char *start, int do_tags,
+         int (*fun) (void *, size_t, const char *, char *, size_t, size_t),
+	 void *closure);
 
 #endif /* not INFO_DISPLAY_H */
